/*
 * Things Gateway App.
 *
 * Back end main script.
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

// Set up the user profile.
const UserProfile = require('./user-profile');
UserProfile.init();
const migration = UserProfile.migrate();

// External Dependencies
const https = require('https');
const http = require('http');
const fs = require('fs');
const express = require('express');
const expressWs = require('express-ws');
const fileUpload = require('express-fileupload');
const bodyParser = require('body-parser');
const GetOpt = require('node-getopt');
const config = require('config');
const path = require('path');
const expressHandlebars = require('express-handlebars');
const ipRegex = require('ip-regex');

// Internal Dependencies
const addonManager = require('./addon-manager');
const db = require('./db');
const mDNSserver = require('./mdns-server');
const Router = require('./router');
const TunnelService = require('./ssltunnel');
const Constants = require('./constants');
const {wifi, wifiSetupApp} = require('./wifi-setup');

// Causes a timestamp to be prepended to console log lines.
require('./log-timestamps');

// The following causes an instance of AppInstance to be created.
// This is then used in other places (like src/addons/plugin/ipc.js)
require('./app-instance');

// Open the database
db.open();

const httpServer = http.createServer();
const httpApp = createGatewayApp(httpServer);

let httpsServer = createHttpsServer();
let httpsApp = null;

/**
 * Creates an HTTPS server object, if successful. If there are no public and
 * private keys stored for the tunnel service, null is returned.
 *
 * @param {}
 * @return {Object|null} https server object if successful, else NULL
 */
function createHttpsServer() {
  if (!TunnelService.hasCertificates()) {
    return null;
  }

  // HTTPS server configuration
  const options = {
    key: fs.readFileSync(path.join(UserProfile.sslDir, 'privatekey.pem')),
    cert: fs.readFileSync(path.join(UserProfile.sslDir, 'certificate.pem')),
  };
  if (fs.existsSync(path.join(UserProfile.sslDir, 'chain.pem'))) {
    options.ca = fs.readFileSync(path.join(UserProfile.sslDir, 'chain.pem'));
  }
  return https.createServer(options);
}

function startHttpsGateway() {
  let port = config.get('ports.https');
  const cliOptions = getOptions();
  if (typeof cliOptions.port === 'number') {
    port = cliOptions.port;
  }

  if (!httpsServer) {
    httpsServer = createHttpsServer();
  }

  httpsApp = createGatewayApp(httpsServer);
  httpsServer.on('request', httpsApp);

  // Start the HTTPS server
  httpsServer.listen(port, function() {
    migration.then(function() {
      addonManager.loadAddons();
    });
    rulesEngineConfigure(httpsServer);
    console.log('HTTPS server listening on port', httpsServer.address().port);
  });

  // Redirect HTTP to HTTPS
  httpServer.on('request', createRedirectApp(httpsServer.address().port));
  const httpPort = config.get('ports.http');
  httpServer.listen(httpPort, function() {
    console.log('Redirector listening on port', httpServer.address().port);
  });
}

function startHttpGateway() {
  httpServer.on('request', httpApp);

  let port = config.get('ports.http');
  const options = getOptions();
  if (typeof options.port === 'number') {
    port = options.port;
  }

  httpServer.listen(port, function() {
    migration.then(function() {
      addonManager.loadAddons();
    });
    rulesEngineConfigure(httpServer);
    console.log('HTTP server listening on port', httpServer.address().port);
  });
}

function stopHttpGateway() {
  httpServer.removeListener('request', httpApp);
}

function startWiFiSetup() {
  httpServer.on('request', wifiSetupApp.onRequest);

  let port = config.get('ports.http');
  const options = getOptions();
  if (typeof options.port === 'number') {
    port = options.port;
  }

  httpServer.listen(port);
}

function stopWiFiSetup() {
  httpServer.removeListener('request', wifiSetupApp.onRequest);
}

function getOptions() {
  if (!config.get('cli')) {
    return {
      debug: false,
      port: null,
    };
  }

  // Command line arguments
  const getopt = new GetOpt([
    ['d', 'debug', 'Enable debug features'],
    ['p', 'port=PORT', 'Specify the server port to use'],
    ['h', 'help', 'Display help' ],
    ['v', 'verbose', 'Show verbose output'],
    ['', 'check-wifi',
     'Run a connection check on the WiFi (only supported on RasPi)'],
  ]);

  const opt = getopt.parseSystem();
  const options = {
    debug: !!opt.options.debug, // cast to bool
    verbose: opt.options.verbose,
  };

  if (opt.options.verbose) {
    console.log(opt);
  }

  if (opt.options.help) {
    getopt.showHelp();
    process.exit(1);
  }

  if (opt.options.port) {
    options.port = parseInt(opt.options.port);
  }

  if (opt.options['check-wifi']) {
    options['check-wifi'] = opt.options['check-wifi'];
  }
  return options;
}

/**
 * Because the rules engine talks to the server over the public HTTP/WS API,
 * the gateway needs to configure it with a JWT and a server address
 * @param {http.Server|https.Server} server
 */
function rulesEngineConfigure(server) {
  const rulesEngine = require('./rules-engine/index.js');
  let protocol = 'https';
  if (server instanceof http.Server) {
    protocol = 'http';
  }
  const gatewayHref = `${protocol}://127.0.0.1:${server.address().port}`;
  rulesEngine.configure(gatewayHref);
}

function createApp() {
  const app = express();
  app.engine('handlebars', expressHandlebars());
  app.set('view engine', 'handlebars');
  app.set('views', Constants.VIEWS_PATH);

  // Use bodyParser to access the body of requests
  app.use(bodyParser.urlencoded({
    extended: false,
  }));
  app.use(bodyParser.json({limit: '1mb'}));

  // Use fileUpload to handle multi-part uploads
  app.use(fileUpload());

  return app;
}

/**
 * @param {http.Server|https.Server} server
 * @return {express.Router}
 */
function createGatewayApp(server) {
  const app = createApp();
  const opt = getOptions();

  // Inject WebSocket support
  expressWs(app, server);

  // Configure router with configured app and command line options.
  Router.configure(app, opt);
  return app;
}

function createRedirectApp(port) {
  const app = createApp();

  // Allow LE challenges, used when renewing domain.
  app.use(
    /^\/\.well-known\/acme-challenge\/.*/,
    function(request, response, next) {
      if (request.method !== 'GET') {
        response.sendStatus(403);
        return;
      }

      const reqPath = path.join(Constants.BUILD_STATIC_PATH, request.path);
      if (fs.existsSync(reqPath)) {
        response.sendFile(reqPath);
        return;
      }

      next();
    });

  // Redirect based on https://https.cio.gov/apis/
  app.use(function(request, response) {
    if (request.method !== 'GET') {
      response.sendStatus(403);
      return;
    }
    if (request.headers.authorization) {
      response.sendStatus(403);
      return;
    }
    let httpsUrl = `https://${request.hostname}`;
    // If we're behind forwarding we can redirect to the port-free https url
    if (port !== 443 && !config.get('behindForwarding')) {
      httpsUrl += `:${port}`;
    }
    httpsUrl += request.url;

    // If the request is for a bare hostname, a .local address, or an IP
    // address, use a 307 redirect to prevent caching. For instance, if the
    // browser caches a redirect for gateway.local to the HTTPS version, things
    // will break after resetting/reflashing the gateway.
    //
    // Otherwise, use a 301 to help mitigate DNS hijacking.
    if (request.hostname.indexOf('.') < 0 ||
        request.hostname.endsWith('.local') ||
        ipRegex({exact: true}).test(request.hostname)) {
      response.redirect(307, httpsUrl);
    } else {
      response.redirect(301, httpsUrl);
    }
  });

  return app;
}

let serverStartup = Promise.resolve();
let wifiPromise = Promise.resolve(true);
const options = getOptions();

if (options['check-wifi']) {
  wifiPromise = wifi.checkConnection();
}

wifiPromise.then((connected) => {
  if (!connected) {
    wifiSetupApp.onConnection = function() {
      stopWiFiSetup();
      startGateway();
    };
    startWiFiSetup();
  } else {
    startGateway();
  }
});

function startGateway() {
  // if we have the certificates installed, we start https
  if (TunnelService.hasCertificates()) {
    serverStartup = TunnelService.userSkipped().then(function(res) {
      if (res) {
        startHttpGateway();
      } else {
        startHttpsGateway();
        TunnelService.hasTunnelToken().then(function(result) {
          if (result) {
            TunnelService.start();
          }
        });
      }
    });
  } else {
    startHttpGateway();
  }
}

if (config.get('cli')) {
  // Get some decent error messages for unhandled rejections. This is
  // often just errors in the code.
  process.on('unhandledRejection', (reason) => {
    console.log('Unhandled Rejection');
    console.error(reason);
  });

  // Do graceful shutdown when Control-C is pressed.
  process.on('SIGINT', function() {
    console.log('Control-C: unloading add-ons...');
    addonManager.unloadAddons();
    mDNSserver.server.cleanup();
    TunnelService.stop();
    process.exit(0);
  });
}

// function to stop running server and start https
TunnelService.switchToHttps = function() {
  stopHttpGateway();
  startHttpsGateway();
};

// This part starts our Service Discovery process.
// We check to see if mDNS should be setup in default mode, or has a previous
// user setup a unique domain. Then we start it.
mDNSserver.getmDNSstate().then((state) => {
  try {
    mDNSserver.getmDNSconfig().then((mDNSconfig) => {
      console.log(`DNS config is: ${mDNSconfig.host}`);
      mDNSserver.server.changeProfile(mDNSconfig);
      mDNSserver.server.setState(state);
    });
  } catch (err) {
    // if we failed to startup mDNS it's not the end of the world log it
    // and carry on
    console.error(`Service Discover failed to start with error: ${err}`);
  }
});

module.exports = { // for testing
  httpServer: httpServer,
  server: httpsServer,
  serverStartup: serverStartup,
};
